/* 
   'ls' for cadaver
   Copyright (C) 2000, Joe Orton <joe@orton.demon.co.uk>, 
   except where otherwise indicated.
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   $Id: ls.c,v 1.10 2000/09/11 16:41:25 joe Exp $
*/

#include <config.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <time.h>

#include <http_request.h>
#include <dav_props.h>
#include <uri.h>
#include <ne_alloc.h>

#include "i18n.h"
#include "commands.h"
#include "cadaver.h"
#include "basename.h"

static time_t current_time;

struct fetch_context {
    struct resource **list;
    const char *target; /* Request-URI of the PROPFIND */
    unsigned int include_target; /* Include resource at href */
};    

static char *format_time(time_t when)
{
    const char *fmt;
    static char ret[256];
    struct tm *local;

    if (when == (time_t)-1) {
	/* Happens on lock-null resources */
	return "  (unknown) ";
    }

    /* from GNU fileutils... this section is 
     *  Copyright (C) 85, 88, 90, 91, 1995-1999 Free Software Foundation, Inc.
     */
    if (current_time > when + 6L * 30L * 24L * 60L * 60L	/* Old. */
	|| current_time < when - 60L * 60L) {
	/* The file is fairly old or in the future.
	   POSIX says the cutoff is 6 months old;
	   approximate this by 6*30 days.
	   Allow a 1 hour slop factor for what is considered "the future",
	   to allow for NFS server/client clock disagreement.
	   Show the year instead of the time of day.  */
	fmt = "%b %e  %Y";
    } else {
	fmt = "%b %e %H:%M";
    }
    /* end FSF copyrighted section. */
    local = localtime(&when);
    if (local != NULL) {
	if (strftime(ret, 256, fmt, local)) {
	    return ret;
	}
    }
    return "???";
}

static int compare_resource(const struct resource *r1, 
			    const struct resource *r2)
{
    /* Sort errors first, then collections, then alphabetically */
    if (r1->type == resr_error) {
	return -1;
    } else if (r2->type == resr_error) {
	return 1;
    } else if (r1->type == resr_collection) {
	if (r2->type != resr_collection) {
	    return -1;
	} else {
	    return strcmp(r1->uri, r2->uri);
	}
    } else {
	if (r2->type != resr_collection) {
	    return strcmp(r1->uri, r2->uri);
	} else {
	    return 1;
	}
    }
}

static void display_ls_line(struct resource *res)
{
    const char *restype;
    char exec_char, *name;

    switch (res->type) {
    case resr_normal: restype = ""; break;
    case resr_reference: restype = _("Ref:"); break;
    case resr_collection: restype = _("Coll:"); break;
    default:
	restype = "???"; break;
    }
    
    if (uri_has_trailing_slash(res->uri)) {
	res->uri[strlen(res->uri)-1] = '\0';
    }
    name = strrchr(res->uri, '/');
    if (name != NULL && strlen(name+1) > 0) {
	name++;
    } else {
	name = res->uri;
    }

    if (res->type == resr_error) {
	printf(_("Error: %-30s %d %s\n"), name, res->error_status,
	       res->error_reason?res->error_reason:_("unknown"));
    } else {
	exec_char = res->is_executable ? '*' : ' ';
	printf("%5s %c%-30s %10d  %s\n", restype, exec_char, name,
	       res->size, format_time(res->modtime));
    }
}

void execute_ls(const char *remote)
{
    int ret;
    char *real_remote;
    struct resource *reslist = NULL, *current, *next;
    if (remote != NULL) {
	real_remote = resolve_path(path, remote, false);
    } else {
	real_remote = ne_strdup(path);
    }
    out_start(_("Listing collection"), real_remote);
    ret = fetch_resource_list(session, real_remote, 1, 0, &reslist);
    if (ret == HTTP_OK) {
	/* Easy this, eh? */
	if (reslist == NULL) {
	    output(o_finish, _("collection is empty.\n"));
	} else {
	    out_success();
	    current_time = time(NULL);
	    for (current = reslist; current!=NULL; current = next) {
		next = current->next;
		if (strlen(current->uri) > strlen(real_remote)) {
		    display_ls_line(current);
		}
		free_resource(current);
	    }
	}
    } else {
	out_result(ret);
    }
    free(real_remote);
}

static void *start_resource(void *userdata, const char *href)
{
    struct fetch_context *ctx = userdata;
    struct resource *res;
    const char *abspath;

    DEBUG(DEBUG_HTTP, "Href: %s\n", href);

    abspath = uri_abspath(href);

    DEBUG(DEBUG_HTTP, "Abspath %s in %s\n", abspath, ctx->target);
    
    if (uri_compare(ctx->target, abspath) == 0 && !ctx->include_target) {
	/* This is the target URI */
	DEBUG(DEBUG_HTTP, "Skipping target resource.\n");
	return NULL;
    }
    
    res = ne_calloc(sizeof(struct resource));
    res->uri = ne_strdup(abspath);
    res->type = resr_normal;
    return res;
}

static void end_resource(void *userdata, void *resource,
			 const char *status_line, const http_status *status,
			 const char *description)
{
    struct fetch_context *ctx = userdata;
    struct resource *current, *previous, *newres = resource;

    if (newres == NULL) {
	return;
    }
    
    if (status && status->_class != 2) {
	/* It's an error! */
	newres->error_status = status->code;
	if (strcmp(status->reason_phrase, "status text goes here") == 0) {
	    const char *desc;
	    /* Special hack for Apache 1.3/mod_dav */
	    if (status->code == 401) {
		desc = _("Authorization Required");
	    } else if (status->_class == 3) {
		desc = _("Redirect");
	    } else if (status->_class == 5) {
		desc = _("Server Error");
	    } else {
		desc = _("Unknown Error");
	    }
	    newres->error_reason = ne_strdup(desc);
	} else {
	    newres->error_reason = ne_strdup(status->reason_phrase);
	}
	newres->type = resr_error;
    }
    
    DEBUG(DEBUG_HTTP, "End resource %s\n", newres->uri);

    for (current = *ctx->list, previous = NULL; current != NULL; 
	 previous = current, current=current->next) {
	if (compare_resource(current, newres) >= 0) {
	    break;
	}
    }
    if (previous) {
	previous->next = newres;
    } else {
	*ctx->list = newres;
    }
    newres->next = current;
}

#define ELM_getcontentlength (1001)
#define ELM_resourcetype (1002)
#define ELM_getlastmodified (1003)
#define ELM_executable (1004)
#define ELM_collection (1005)
#define ELM_displayname (1006)

static int end_element(void *userdata, const struct hip_xml_elm *elm, const char *cdata)
{
    struct resource *r;
    r = dav_propfind_get_current_resource(userdata);
    if (r == NULL) {
	return 0;
    }
    DEBUG(DEBUG_HTTP, "In resource %s\n", r->uri);
    DEBUG(DEBUG_HTTP, "Property [%d], %s@@%s = %s\n", elm->id,
	   elm->nspace, elm->name, cdata?cdata:"undefined");
    switch (elm->id) {
    case ELM_collection:
	r->type = resr_collection;
	break;
    case ELM_getcontentlength:
	if (cdata) {
	    r->size = atoi(cdata);
	}
	break;
    case ELM_displayname:
	if (cdata) {
	    r->displayname = ne_strdup(cdata);
	}
	break;
    case ELM_getlastmodified:
	if (cdata) {
	    r->modtime = http_dateparse(cdata);
	    /* will be -1 on error */
	}
	break;
    case ELM_executable:
	if (cdata) {
	    if (strcasecmp(cdata, "T") == 0) {
		r->is_executable = 1;
	    } else {
		r->is_executable = 0;
	    }
	}
	break;
    }
    return 0;
}

static int check_context(hip_xml_elmid child, hip_xml_elmid parent)
{
    /* FIXME */
    return 0;    
}

void free_resource(struct resource *res)
{
    HTTP_FREE(res->uri);
    HTTP_FREE(res->displayname);
    HTTP_FREE(res->error_reason);
    free(res);
}

void free_resource_list(struct resource *res)
{
    struct resource *next;
    for (; res != NULL; res = next) {
	next = res->next;
	free_resource(res);
    }
}

int fetch_resource_list(http_session *sess, const char *uri,
			 int depth, int include_target,
			 struct resource **reslist)
{
    dav_propfind_handler *ph = dav_propfind_create(sess, uri, depth);
    int ret;
    static const dav_propname props[] = {
	{ "DAV:", "getcontentlength" },
	{ "DAV:", "resourcetype" },
	{ "DAV:", "getlastmodified" },
	{ "DAV:", "displayname" },
	{ "http://apache.org/dav/props/", "executable" },
	{ NULL }
    };
    static const struct hip_xml_elm elms[] = {
	{ "DAV:", "getcontentlength", ELM_getcontentlength, HIP_XML_CDATA },
	{ "DAV:", "resourcetype", ELM_resourcetype, 0 },
	{ "DAV:", "getlastmodified", ELM_getlastmodified, HIP_XML_CDATA },
	{ "http://apache.org/dav/props/", "executable", ELM_executable, HIP_XML_CDATA },
	{ "DAV:", "collection", ELM_collection, 0 },
	{ "DAV:", "displayname", ELM_displayname, HIP_XML_CDATA },
	{ NULL }
    };
    struct fetch_context ctx = {0};
    
    dav_propfind_set_resource_handlers(ph, start_resource, end_resource);

    /* Note that the handler itself (not the context) is passed as
     * userdata to the callbacks here.  This is so the current
     * resource cookie can be retrieved. */
    hip_xml_add_handler(dav_propfind_get_parser(ph),
			elms, check_context, NULL, end_element, ph);
    
    *reslist = NULL;
    ctx.list = reslist;
    ctx.target = uri;
    ctx.include_target = include_target;

    ret = dav_propfind_named(ph, props, &ctx);

    dav_propfind_destroy(ph);

    return ret;
}
